"""Implementation of Property List Encoders
"""

from . import XML_FORMAT, BINARY_FORMAT

from abc import ABCMeta, abstractmethod
import plistlib
import types
import struct
import datetime


class PlistEncoder(object):
    __metaclass__ = ABCMeta

    def __new__(cls, fp, format=XML_FORMAT):
        if cls is PlistEncoder:
            if format == XML_FORMAT:
                return super(PlistEncoder, cls).__new__(_XMLPlistEncoder)
            elif format == BINARY_FORMAT:
                return super(PlistEncoder, cls).__new__(_BinaryPlistEncoder)
            else:
                raise NotImplementedError('Unknown format.')
        else:
            return super(PlistEncoder, cls).__new__(cls, fp, format)

    def __init__(self, fp, format=XML_FORMAT):
        self.write = fp.write

    @abstractmethod
    def dump(self, obj):
        pass


class _XMLPlistEncoder(PlistEncoder):
    def __init__(self, fp, format=XML_FORMAT):
        super(_XMLPlistEncoder, self).__init__(fp, format)

    def dump(self, obj):
        # fall back on plistlib until we have implemented this ourselves
        self.write(plistlib.writePlistToString(obj))


# Markers for the binary format
M_SIMPLE = 0x00
M_NULL   = 0x00
M_FALSE  = 0x08
M_TRUE   = 0x09
M_FILL   = 0x0f

M_INT    = 0x10
M_REAL   = 0x20
M_DATE   = 0x30
M_DATA   = 0x40
M_STRING = 0x50
M_USTRING = 0x60
M_UID    = 0x80
M_ARRAY  = 0xa0
M_SET    = 0xc0
M_DICT   = 0xd0


# Apple uses 1 January 2001 as the reference date
REFERENCE_DATE = datetime.datetime(2001, 1, 1)


class _BinaryPlistEncoder(PlistEncoder):
    """Binary Property List encoder for Python data structures.

    The format is documented in

        http://opensource.apple.com/source/CF/CF-550/CFBinaryPList.c


    HEADER
        magic number ("bplist")
        file format version

    OBJECT TABLE
        variable-sized objects

        Object Formats (marker byte followed by additional info in some cases)
        null    0000 0000
        bool    0000 1000                  // false
        bool    0000 1001                  // true
        fill    0000 1111                  // fill byte
        int     0001 nnnn  ...             // # of bytes is 2^nnnn, big-endian bytes
        real    0010 nnnn  ...             // # of bytes is 2^nnnn, big-endian bytes
        date    0011 0011  ...             // 8 byte float follows, big-endian bytes
        data    0100 nnnn  [int]   ...     // nnnn is number of bytes unless 1111 then int count follows, followed by bytes
        string  0101 nnnn  [int]   ...     // ASCII string, nnnn is # of chars, else 1111 then int count, then bytes
        string  0110 nnnn  [int]   ...     // Unicode string, nnnn is # of chars, else 1111 then int count, then big-endian
                                           // 2-byte uint16_t
                0111 xxxx                  // unused
        uid     1000 nnnn  ...             // nnnn+1 is # of bytes
                1001 xxxx                  // unused
        array   1010 nnnn  [int]   objref* // nnnn is count, unless '1111', then int count follows
                1011 xxxx                  // unused
        set     1100 nnnn  [int]   objref* // nnnn is count, unless '1111', then int count follows
        dict    1101 nnnn  [int]   keyref* objref* // nnnn is count, unless '1111', then int count follows
                1110 xxxx                  // unused
                1111 xxxx                  // unused

    OFFSET TABLE
        list of ints, byte size of which is given in trailer
        -- these are the byte offsets into the file
        -- number of these is in the trailer

    TRAILER
        byte size of offset ints in offset table
        byte size of object refs in arrays and dicts
        number of offsets in offset table (also is number of objects)
        element # in offset table which is top level object
        offset table offset
    """

    def __init__(self, fp, format=BINARY_FORMAT):
        super(_BinaryPlistEncoder, self).__init__(fp, format)

        self.offset_table = []

        # trailer info (object count == len(offset_table)
        self.offset_size = 1
        self.ref_size = 1
        self.top_level_object_ref = 0
        self.offset_table_offset = 0

    def dump(self, obj):
        """Serialize ``obj`` as a Property List formatted stream to ``fp`` (a
        file-like object).
        """
        self.write(self.enc_header())
        ref, object = self.enc_object_table(obj)
        self.top_level_object_ref = ref
        self.write(object)
        self.write(self.enc_offset_table())
        self.write(self.enc_trailer())

    def enc_header(self):
        self.offset_table_offset += 8
        magic_number = 'bplist'
        version = '00'
        return magic_number + version

    def enc_object_table(self, obj):
        t = type(obj)
        f = self.dispatch.get(t)

        ref = len(self.offset_table)
        self.offset_table.append(self.offset_table_offset)

        if f:
            return ref, f(self, obj)     # unbound method

        raise TypeError("Don't know how to encode %s object" % t)

    def enc_offset_table(self):
        refs = [struct.pack('B', offset) for offset in self.offset_table]
        return ''.join(refs)

    def enc_trailer(self):
        trailer = struct.pack('!xxxxxxBBQQQ', self.offset_size, self.ref_size,
                              len(self.offset_table), self.top_level_object_ref,
                              self.offset_table_offset)
        return trailer

    # Methods below are dispatched through the dispatch table

    dispatch = {}

    def enc_simple(self, payload):
        self.offset_table_offset += 1
        return struct.pack('>B', M_SIMPLE | payload)

    def enc_none(self, obj):
        # Null values seem to be supported by the binary format but not by
        # the xml format - at least plutil -lint doesn't complain.
        payload = M_NULL
        return self.enc_simple(payload)
    dispatch[types.NoneType] = enc_none

    def enc_bool(self, obj):
        payload = M_TRUE if obj else M_FALSE
        return self.enc_simple(payload)
    dispatch[types.BooleanType] = enc_bool

    def enc_int(self, obj):
        # CFBinaryPlist.c:
        # in format version '00', 1, 2, and 4-byte integers have to be
        # interpreted as unsigned, whereas 8-byte integers are signed (and
        # 16-byte when available)
        # negative 1, 2, 4-byte integers are always emitted as 8 bytes in
        # format '00'
        # integers are not required to be in the most compact possible
        # representation, but only the last 64 bits are significant currently
        if obj < 0:
            size = 3
            type = 'q'
        elif obj < 2**8:
            size = 0
            type = 'B'
        elif obj < 2**16:
            size = 1
            type = 'H'
        elif obj < 2**32:
            size = 2
            type = 'L'
        elif obj < 2**63-1:     # MAXINT
            size = 3
            type = 'q'
        else:
            raise ValueError("int too big.")

        self.offset_table_offset += 1 + 2**size
        return struct.pack('>B%s' % type, M_INT | size, obj)
    dispatch[types.IntType] = enc_int

    def enc_float(self, obj):
        size = 3
        self.offset_table_offset += 1 + 2**size
        return struct.pack('>Bd', M_REAL | size, obj)
    dispatch[types.FloatType] = enc_float

    def enc_date(self, obj):
        # convert to naive datetime object so we can subtract
        if obj.tzinfo:
            obj = obj.replace(tzinfo=None) - obj.utcoffset()
        timedelta = obj - REFERENCE_DATE
        seconds = timedelta.total_seconds()

        size = 3
        self.offset_table_offset += 1 + 2**size
        return struct.pack('>Bd', M_DATE | size, seconds)
    dispatch[datetime.datetime] = enc_date

    def enc_object_with_count(self, marker, count):
        # common structure for most types
        if count > 14:
            self.offset_table_offset += 1
            return struct.pack('>B', marker | 0x0f) + self.enc_int(count)
        else:
            self.offset_table_offset += 1
            return struct.pack('>B', marker | count)

    def enc_string(self, obj):
        count = len(obj)
        payload = obj
        self.offset_table_offset += len(payload)
        return self.enc_object_with_count(M_STRING, count) + payload
    dispatch[types.StringType] = enc_string

    def enc_unicode(self, obj):
        count = len(obj)
        payload = obj.encode('UTF-16BE', 'strict')
        self.offset_table_offset += len(payload)
        return self.enc_object_with_count(M_USTRING, count) + payload
    dispatch[types.UnicodeType] = enc_unicode

    def enc_list(self, obj):
        count = len(obj)
        objects = []
        object_refs = []

        wrapper = self.enc_object_with_count(M_ARRAY, count)
        self.offset_table_offset += count

        for element in obj:
            ref, object = self.enc_object_table(element)
            objects.append(object)
            object_ref = struct.pack('>B', ref)
            object_refs.append(object_ref)

        return wrapper + ''.join(object_refs) + ''.join(objects)
    dispatch[types.ListType] = enc_list

    def enc_dict(self, obj):
        count = len(obj)
        keys, key_refs = [], []
        objects, object_refs = [], []

        wrapper = self.enc_object_with_count(M_DICT, count)
        self.offset_table_offset += 2*count

        for key in obj.keys():
            ref, object = self.enc_object_table(str(key))
            keys.append(object)
            key_ref = struct.pack('>B', ref)
            key_refs.append(key_ref)

        for value in obj.values():
            ref, object = self.enc_object_table(value)
            objects.append(object)
            object_ref = struct.pack('>B', ref)
            object_refs.append(object_ref)

        return (wrapper + ''.join(key_refs) + ''.join(object_refs) +
                ''.join(keys) + ''.join(objects))
    dispatch[types.DictType] = enc_dict
